import enum
import functools
import sys
import textwrap
import traceback

from click import echo, style

from ..exceptions import BackendAPIError

__all__ = (
    "PrintStatus",
    "print_pretty",
    "print_info",
    "print_wait",
    "print_done",
    "print_warn",
    "print_fail",
    "print_error",
    "show_warning",
)


class PrintStatus(enum.Enum):
    NONE = 0
    WAITING = 1
    DONE = 2
    FAILED = 3
    WARNING = 4


def bold(text: str) -> str:
    """
    Wraps the given text with bold enable/disable ANSI sequences.
    """
    return style(text, bold=True, reset=False) + style("", bold=False, reset=False)


def underline(text: str) -> str:
    return style(text, underline=True, reset=False) + style("", underline=False, reset=False)


def inverse(text: str) -> str:
    return style(text, reverse=True, reset=False) + style("", reverse=False, reset=False)


def italic(text: str) -> str:
    return "\x1b[3m" + text + "\x1b[23m"


def format_pretty(msg, status=PrintStatus.NONE, colored=True):
    if status == PrintStatus.NONE:
        indicator = style("\u2219", fg="bright_cyan", reset=False)
    elif status == PrintStatus.WAITING:
        indicator = style("\u22EF", fg="bright_yellow", reset=False)
    elif status == PrintStatus.DONE:
        indicator = style("\u2714", fg="bright_green", reset=False)
    elif status == PrintStatus.FAILED:
        indicator = style("\u2718", fg="bright_red", reset=False)
    elif status == PrintStatus.WARNING:
        indicator = style("\u2219", fg="yellow", reset=False)
    else:
        raise ValueError
    return style(indicator + textwrap.indent(msg, "  ")[1:], reset=True)


format_info = functools.partial(format_pretty, status=PrintStatus.NONE)
format_wait = functools.partial(format_pretty, status=PrintStatus.WAITING)
format_done = functools.partial(format_pretty, status=PrintStatus.DONE)
format_fail = functools.partial(format_pretty, status=PrintStatus.FAILED)
format_warn = functools.partial(format_pretty, status=PrintStatus.WARNING)


def print_pretty(msg, *, status=PrintStatus.NONE, file=None):
    if file is None:
        file = sys.stderr
    if status == PrintStatus.NONE:
        indicator = style("\u2219", fg="bright_cyan", reset=False)
    elif status == PrintStatus.WAITING:
        assert "\n" not in msg, "Waiting message must be a single line."
        indicator = style("\u22EF", fg="bright_yellow", reset=False)
    elif status == PrintStatus.DONE:
        indicator = style("\u2714", fg="bright_green", reset=False)
    elif status == PrintStatus.FAILED:
        indicator = style("\u2718", fg="bright_red", reset=False)
    elif status == PrintStatus.WARNING:
        indicator = style("\u2219", fg="yellow", reset=False)
    else:
        raise ValueError
    echo("\x1b[2K", nl=False, file=file)
    text = textwrap.indent(msg, "  ")
    text = style(indicator + text[1:], reset=True)
    echo("{0}\r".format(text), nl=False, file=file)
    file.flush()
    if status != PrintStatus.WAITING:
        echo("", file=file)


print_info = functools.partial(print_pretty, status=PrintStatus.NONE)
print_wait = functools.partial(print_pretty, status=PrintStatus.WAITING)
print_done = functools.partial(print_pretty, status=PrintStatus.DONE)
print_fail = functools.partial(print_pretty, status=PrintStatus.FAILED)
print_warn = functools.partial(print_pretty, status=PrintStatus.WARNING)


def format_error(exc: Exception):
    if isinstance(exc, BackendAPIError):
        yield "{0}: {1} {2}\n".format(exc.__class__.__name__, exc.status, exc.reason)
        yield "{0[title]}".format(exc.data)
        if exc.data["type"].endswith("/too-many-sessions-matched"):
            matches = exc.data["data"].get("matches", [])
            if matches:
                yield "\nCandidates (up to 10 recent entries):\n"
            for item in matches:
                yield f"- {item['id']} ({item['name']}, {item['status']})\n"
        elif exc.data["type"].endswith("/session-already-exists"):
            existing_session_id = exc.data["data"].get("existingSessionId", None)
            if existing_session_id is not None:
                yield f"\n- Existing session ID: {existing_session_id}"
        elif exc.data["type"].endswith("/invalid-api-params"):
            general_error_msg = exc.data.get("msg", None)
            if general_error_msg is not None:
                yield f"\n- {general_error_msg}"
            per_field_errors = exc.data.get("data", {})
            if isinstance(per_field_errors, dict):
                for k, v in per_field_errors.items():
                    yield f'\n- "{k}": {v}'
            else:
                yield f"\n- {per_field_errors}"
        else:
            if exc.data["type"].endswith("/graphql-error"):
                yield "\n\u279c Message:\n"
                yield from (f"{err_item['message']}\n" for err_item in exc.data.get("data", []))
            else:
                other_details = exc.data.get("msg", None)
                if other_details:
                    yield "\n\u279c Message: "
                    yield str(other_details)
                other_data = exc.data.get("data", None)
                if other_data:
                    yield "\n\u279c Data: "
                    yield repr(other_data)
        agent_details = exc.data.get("agent-details", None)
        if agent_details is not None:
            yield "\n\u279c This is an agent-side error. "
            yield "Check the agent status or ask the administrator for help."
            agent_exc = agent_details.get("exception", None)
            if agent_exc is not None:
                yield "\n\u279c " + str(agent_exc)
            desc = agent_details.get("title", None)
            if desc is not None:
                yield "\n\u279c " + str(desc)
        content = exc.data.get("content", None)
        if content:
            yield "\n" + content
    else:
        args = exc.args if exc.args else [""]
        yield f"{exc.__class__.__name__}: {args[0]}\n"
        yield "\n".join(map(str, args[1:]))
        yield ("*** Traceback ***\n" + "".join(traceback.format_tb(exc.__traceback__)).strip())


def print_error(exc: Exception, *, file=None):
    if file is None:
        file = sys.stderr
    indicator = style("\u2718", fg="bright_red", reset=False)
    if file.isatty():
        echo("\x1b[2K", nl=False, file=file)
    text = "".join(format_error(exc))
    text = textwrap.indent(text, "  ")
    text = style(indicator + text[1:], reset=True)
    echo("{0}\r".format(text), nl=False, file=file)
    echo("", file=file)
    file.flush()


def show_warning(message, category, filename, lineno, file=None, line=None):
    echo(
        "{0}: {1}".format(
            style(str(category.__name__), fg="yellow", bold=True),
            style(str(message), fg="yellow"),
        ),
        file=file,
    )



import sys

import click

from ai.backend.client.func.scaling_group import _default_detail_fields, _default_list_fields
from ai.backend.client.output.fields import scaling_group_fields
from ai.backend.client.session import Session

from ..extensions import pass_ctx_obj
from ..params import JSONParamType
from ..types import CLIContext
from . import admin


@admin.group()
def scaling_group() -> None:
    """
    Scaling group (resource group) administration commands.
    """


@scaling_group.command()
@pass_ctx_obj
@click.argument("group", type=str, metavar="GROUP_NAME")
def get_available(ctx: CLIContext, group: str) -> None:
    with Session() as session:
        try:
            items = session.ScalingGroup.list_available(group)
            ctx.output.print_list(items, [scaling_group_fields["name"]])
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(1)


@scaling_group.command()
@pass_ctx_obj
@click.argument("name", type=str)
def info(ctx: CLIContext, name: str) -> None:
    """
    Show the information about the given scaling group.
    (superadmin privilege required)
    """
    with Session() as session:
        try:
            item = session.ScalingGroup.detail(name=name)
            ctx.output.print_item(item, _default_detail_fields)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(1)


@scaling_group.command()
@pass_ctx_obj
def list(ctx: CLIContext) -> None:
    """
    List and manage scaling groups.
    (superadmin privilege required)
    """
    with Session() as session:
        try:
            items = session.ScalingGroup.list()
            ctx.output.print_list(items, _default_list_fields)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(1)


@scaling_group.command()
@pass_ctx_obj
@click.argument("name", type=str, metavar="NAME")
@click.option("-d", "--description", type=str, default="", help="Description of new scaling group")
@click.option("-i", "--inactive", is_flag=True, help="New scaling group will be inactive.")
@click.option("--driver", type=str, default="static", help="Set driver.")
@click.option(
    "--driver-opts", type=JSONParamType(), default="{}", help="Set driver options as a JSON string."
)
@click.option("--scheduler", type=str, default="fifo", help="Set scheduler.")
@click.option(
    "--scheduler-opts",
    type=JSONParamType(),
    default="{}",
    help="Set scheduler options as a JSON string.",
)
def add(
    ctx: CLIContext, name, description, inactive, driver, driver_opts, scheduler, scheduler_opts
):
    """
    Add a new scaling group.

    NAME: Name of new scaling group.
    """
    with Session() as session:
        try:
            data = session.ScalingGroup.create(
                name,
                description=description,
                is_active=not inactive,
                driver=driver,
                driver_opts=driver_opts,
                scheduler=scheduler,
                scheduler_opts=scheduler_opts,
            )
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name="scaling_group",
                action_name="add",
            )
            sys.exit(1)
        if not data["ok"]:
            ctx.output.print_mutation_error(
                msg=data["msg"],
                item_name="scaling_group",
                action_name="add",
            )
            sys.exit(1)
        ctx.output.print_mutation_result(
            data,
            item_name="scaling_group",
        )


@scaling_group.command()
@pass_ctx_obj
@click.argument("name", type=str, metavar="NAME")
@click.option("-d", "--description", type=str, default="", help="Description of new scaling group")
@click.option("-i", "--inactive", is_flag=True, help="New scaling group will be inactive.")
@click.option("--driver", type=str, default="static", help="Set driver.")
@click.option(
    "--driver-opts", type=JSONParamType(), default=None, help="Set driver options as a JSON string."
)
@click.option("--scheduler", type=str, default="fifo", help="Set scheduler.")
@click.option(
    "--scheduler-opts",
    type=JSONParamType(),
    default=None,
    help="Set scheduler options as a JSON string.",
)
def update(
    ctx: CLIContext, name, description, inactive, driver, driver_opts, scheduler, scheduler_opts
):
    """
    Update existing scaling group.

    NAME: Name of new scaling group.
    """
    with Session() as session:
        try:
            data = session.ScalingGroup.update(
                name,
                description=description,
                is_active=not inactive,
                driver=driver,
                driver_opts=driver_opts,
                scheduler=scheduler,
                scheduler_opts=scheduler_opts,
            )
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name="scaling_group",
                action_name="update",
            )
            sys.exit(1)
        if not data["ok"]:
            ctx.output.print_mutation_error(
                msg=data["msg"],
                item_name="scaling_group",
                action_name="update",
            )
            sys.exit(1)
        ctx.output.print_mutation_result(
            data,
            extra_info={
                "name": name,
            },
        )


@scaling_group.command()
@pass_ctx_obj
@click.argument("name", type=str, metavar="NAME")
def delete(ctx: CLIContext, name):
    """
    Delete an existing scaling group.

    NAME: Name of a scaling group to delete.
    """
    with Session() as session:
        try:
            data = session.ScalingGroup.delete(name)
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name="scaling_group",
                action_name="deletion",
            )
            sys.exit(1)
        if not data["ok"]:
            ctx.output.print_mutation_error(
                msg=data["msg"],
                item_name="scaling_group",
                action_name="deletion",
            )
            sys.exit(1)
        ctx.output.print_mutation_result(
            data,
            extra_info={
                "name": name,
            },
        )


@scaling_group.command()
@pass_ctx_obj
@click.argument("scaling_group", type=str, metavar="SCALING_GROUP")
@click.argument("domain", type=str, metavar="DOMAIN")
def associate_scaling_group(ctx: CLIContext, scaling_group, domain):
    """
    Associate a domain with a scaling_group.

    \b
    SCALING_GROUP: The name of a scaling group.
    DOMAIN: The name of a domain.
    """
    with Session() as session:
        try:
            data = session.ScalingGroup.associate_domain(scaling_group, domain)
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name="scaling_group",
                action_name="scaling_group_association",
            )
            sys.exit(1)
        if not data["ok"]:
            ctx.output.print_mutation_error(
                msg=data["msg"],
                item_name="scaling_group",
                action_name="scaling_group_association",
            )
            sys.exit(1)
        ctx.output.print_mutation_result(
            data,
            extra_info={
                "detail_msg": "Scaling group {} is assocatiated with domain {}.".format(
                    scaling_group, domain
                ),
            },
        )


@scaling_group.command()
@pass_ctx_obj
@click.argument("scaling_group", type=str, metavar="SCALING_GROUP")
@click.argument("domain", type=str, metavar="DOMAIN")
def dissociate_scaling_group(ctx: CLIContext, scaling_group, domain):
    """
    Dissociate a domain from a scaling_group.

    \b
    SCALING_GROUP: The name of a scaling group.
    DOMAIN: The name of a domain.
    """
    with Session() as session:
        try:
            data = session.ScalingGroup.dissociate_domain(scaling_group, domain)
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name="scaling_group",
                action_name="scaling_group_dissociation",
            )
            sys.exit(1)
        if not data["ok"]:
            ctx.output.print_mutation_error(
                msg=data["msg"],
                item_name="scaling_group",
                action_name="scaling_group_dissociation",
            )
            sys.exit(1)
        ctx.output.print_mutation_result(
            data,
            item_name="scaling_group",
            extra_info={
                "detail_msg": "Scaling group {} is dissociated from domain {}.".format(
                    scaling_group, domain
                ),
            },
        )



from pathlib import Path

from . import exceptions, session

__all__ = (
    *exceptions.__all__,
    *session.__all__,
)

__version__ = (Path(__file__).parent / "VERSION").read_text().strip()


def get_user_agent():
    return "Backend.AI Client for Python {0}".format(__version__)



from __future__ import annotations

import asyncio
import json
import re
from typing import AsyncIterator, Tuple, Union

import aiohttp
import click
from aiohttp import web

from ai.backend.cli.main import main

from ..exceptions import BackendAPIError, BackendClientError
from ..request import Request
from ..session import AsyncSession
from .pretty import print_error, print_fail, print_info


class WebSocketProxy:
    __slots__ = (
        "up_conn",
        "down_conn",
        "upstream_buffer",
        "upstream_buffer_task",
    )

    upstream_buffer: asyncio.Queue[Tuple[Union[str, bytes], aiohttp.WSMsgType]]

    def __init__(self, up_conn: aiohttp.ClientWebSocketResponse, down_conn: web.WebSocketResponse):
        self.up_conn = up_conn
        self.down_conn = down_conn
        self.upstream_buffer = asyncio.Queue()
        self.upstream_buffer_task = None

    async def proxy(self):
        asyncio.ensure_future(self.downstream())
        await self.upstream()

    async def upstream(self):
        try:
            async for msg in self.down_conn:
                if msg.type in (aiohttp.WSMsgType.TEXT, aiohttp.WSMsgType.BINARY):
                    await self.send(msg.data, msg.type)
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    print_fail(
                        "ws connection closed with exception {}".format(self.up_conn.exception())
                    )
                    break
                elif msg.type == aiohttp.WSMsgType.CLOSE:
                    break
            # here, client gracefully disconnected
        except asyncio.CancelledError:
            # here, client forcibly disconnected
            pass
        finally:
            await self.close_downstream()

    async def downstream(self):
        try:
            self.upstream_buffer_task = asyncio.ensure_future(self.consume_upstream_buffer())
            print_info("websocket proxy started")
            async for msg in self.up_conn:
                if msg.type == aiohttp.WSMsgType.TEXT:
                    await self.down_conn.send_str(msg.data)
                elif msg.type == aiohttp.WSMsgType.BINARY:
                    await self.down_conn.send_bytes(msg.data)
                elif msg.type == aiohttp.WSMsgType.CLOSED:
                    break
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    break
            # here, server gracefully disconnected
        except asyncio.CancelledError:
            pass
        except Exception as e:
            print_fail("unexpected error: {}".format(e))
        finally:
            await self.close_upstream()
            print_info("websocket proxy terminated")

    async def consume_upstream_buffer(self):
        try:
            while True:
                data, tp = await self.upstream_buffer.get()
                if not self.up_conn.closed:
                    if tp == aiohttp.WSMsgType.BINARY:
                        await self.up_conn.send_bytes(data)
                    elif tp == aiohttp.WSMsgType.TEXT:
                        await self.up_conn.send_str(data)
        except asyncio.CancelledError:
            pass

    async def send(self, msg: str, tp: aiohttp.WSMsgType):
        await self.upstream_buffer.put((msg, tp))

    async def close_downstream(self):
        if not self.down_conn.closed:
            await self.down_conn.close()

    async def close_upstream(self):
        if not self.upstream_buffer_task.done():
            self.upstream_buffer_task.cancel()
            await self.upstream_buffer_task
        if not self.up_conn.closed:
            await self.up_conn.close()


def _translate_headers(upstream_request: Request, client_request: Request) -> None:
    for k, v in client_request.headers.items():
        upstream_request.headers[k] = v
    api_endpoint = upstream_request.config.endpoint
    assert api_endpoint.host is not None
    if api_endpoint.is_default_port():
        upstream_request.headers["Host"] = api_endpoint.host
    else:
        upstream_request.headers["Host"] = f"{api_endpoint.host}:{api_endpoint.port}"


async def web_handler(request):
    path = re.sub(r"^/?v(\d+)/", "/", request.path)
    try:
        # We treat all requests and responses as streaming universally
        # to be a transparent proxy.
        api_rqst = Request(
            request.method,
            path,
            request.content,
            params=request.query,
        )
        _translate_headers(api_rqst, request)
        if "Content-Type" in request.headers:
            api_rqst.content_type = request.content_type  # set for signing
        # Uploading request body happens at the entering of the block,
        # and downloading response body happens in the read loop inside.
        async with api_rqst.fetch() as up_resp:
            down_resp = web.StreamResponse()
            down_resp.set_status(up_resp.status, up_resp.reason)
            down_resp.headers.update(up_resp.headers)
            down_resp.headers["Access-Control-Allow-Origin"] = "*"
            await down_resp.prepare(request)
            while True:
                chunk = await up_resp.read(8192)
                if not chunk:
                    break
                await down_resp.write(chunk)
            return down_resp
    except BackendAPIError as e:
        return web.Response(body=json.dumps(e.data), status=e.status, reason=e.reason)
    except BackendClientError:
        return web.Response(
            body="The proxy target server is inaccessible.", status=502, reason="Bad Gateway"
        )
    except asyncio.CancelledError:
        return web.Response(
            body="The proxy is being shut down.", status=503, reason="Service Unavailable"
        )
    except Exception as e:
        print_error(e)
        return web.Response(
            body="Something has gone wrong.", status=500, reason="Internal Server Error"
        )


async def websocket_handler(request):
    path = re.sub(r"^/?v(\d+)/", "/", request.path)
    try:
        api_rqst = Request(
            request.method,
            path,
            request.content,
            params=request.query,
            content_type=request.content_type,
        )
        _translate_headers(api_rqst, request)
        async with api_rqst.connect_websocket() as up_conn:
            down_conn = web.WebSocketResponse()
            await down_conn.prepare(request)
            web_socket_proxy = WebSocketProxy(up_conn, down_conn)
            await web_socket_proxy.proxy()
            return down_conn
    except BackendAPIError as e:
        return web.Response(body=json.dumps(e.data), status=e.status, reason=e.reason)
    except BackendClientError:
        return web.Response(
            body="The proxy target server is inaccessible.", status=502, reason="Bad Gateway"
        )
    except asyncio.CancelledError:
        return web.Response(
            body="The proxy is being shut down.", status=503, reason="Service Unavailable"
        )
    except Exception as e:
        print_error(e)
        return web.Response(
            body="Something has gone wrong.", status=500, reason="Internal Server Error"
        )


async def proxy_context(app: web.Application) -> AsyncIterator[None]:
    app["client_session"] = AsyncSession()
    async with app["client_session"]:
        yield


def create_proxy_app():
    app = web.Application()
    app.cleanup_ctx.append(proxy_context)

    app.router.add_route("GET", r"/stream/{path:.*$}", websocket_handler)
    app.router.add_route("GET", r"/wsproxy/{path:.*$}", websocket_handler)
    app.router.add_route("*", r"/{path:.*$}", web_handler)
    return app


@main.command(context_settings=dict(allow_extra_args=True))
@click.option(
    "--bind", type=str, default="localhost", help="The IP/host address to bind this proxy."
)
@click.option(
    "-p",
    "--port",
    type=int,
    default=8084,
    help="The TCP port to accept non-encrypted non-authorized " "API requests.",
)
@click.pass_context
def proxy(ctx, bind, port):
    """
    Run a non-encrypted non-authorized API proxy server.
    Use this only for development and testing!
    """
    app = create_proxy_app()
    web.run_app(app, host=bind, port=port)



from ..request import Request
from .base import BaseFunction, api_function

__all__ = ("EtcdConfig",)


class EtcdConfig(BaseFunction):
    """
    Provides a way to get or set ETCD configurations.

    .. note::

      All methods in this function class require your API access key to
      have the *superadmin* privilege.
    """

    @api_function
    @classmethod
    async def get(cls, key: str, prefix: bool = False) -> dict:
        """
        Get configuration from ETCD with given key.

        :param key: Name of the key to fetch.
        :param prefix: get all keys prefixed with the give key.
        """
        rqst = Request("POST", "/config/get")
        rqst.set_json(
            {
                "key": key,
                "prefix": prefix,
            }
        )
        async with rqst.fetch() as resp:
            data = await resp.json()
            return data.get("result", None)

    @api_function
    @classmethod
    async def set(cls, key: str, value: str) -> dict:
        """
        Set configuration into ETCD with given key and value.

        :param key: Name of the key to set.
        :param value: Value to set.
        """
        rqst = Request("POST", "/config/set")
        rqst.set_json(
            {
                "key": key,
                "value": value,
            }
        )
        async with rqst.fetch() as resp:
            data = await resp.json()
            return data

    @api_function
    @classmethod
    async def delete(cls, key: str, prefix: bool = False) -> dict:
        """
        Delete configuration from ETCD with given key.

        :param key: Name of the key to delete.
        :param prefix: delete all keys prefixed with the give key.
        """
        rqst = Request("POST", "/config/delete")
        rqst.set_json(
            {
                "key": key,
                "prefix": prefix,
            }
        )
        async with rqst.fetch() as resp:
            data = await resp.json()
            return data



from ..request import Request
from .base import BaseFunction, api_function

__all__ = ("Auth",)


class Auth(BaseFunction):
    """
    Provides the function interface for login session management and authorization.
    """

    @api_function
    @classmethod
    async def login(cls, user_id: str, password: str) -> dict:
        """
        Log-in into the endpoint with the given user ID and password.
        It creates a server-side web session and return
        a dictionary with ``"authenticated"`` boolean field and
        JSON-encoded raw cookie data.
        """
        rqst = Request("POST", "/server/login")
        rqst.set_json(
            {
                "username": user_id,
                "password": password,
            }
        )
        async with rqst.fetch(anonymous=True) as resp:
            data = await resp.json()
            data["cookies"] = resp.raw_response.cookies
            data["config"] = {
                "username": user_id,
            }
            return data

    @api_function
    @classmethod
    async def logout(cls) -> None:
        """
        Log-out from the endpoint.
        It clears the server-side web session.
        """
        rqst = Request("POST", "/server/logout")
        async with rqst.fetch() as resp:
            resp.raw_response.raise_for_status()

    @api_function
    @classmethod
    async def update_password(
        cls, old_password: str, new_password: str, new_password2: str
    ) -> dict:
        """
        Update user's password. This API works only for account owner.
        """
        rqst = Request("POST", "/auth/update-password")
        rqst.set_json(
            {
                "old_password": old_password,
                "new_password": new_password,
                "new_password2": new_password2,
            }
        )
        async with rqst.fetch() as resp:
            return await resp.json()
